"""Module containing vanilla commands"""
COMMANDS = {
    # OP Level 0-2
    "advancement",
    "attribute",
    "bossbar",
    "clear",
    "clone",
    "data",
    "datapack",
    "debug",
    "defaultgamemode",
    "difficulty",
    "effect",
    "enchant",
    "execute",
    "experience",
    "fill",
    "forceload",
    "function",
    "gamemode",
    "gamerule",
    "give",
    "help",
    "item",
    "kill",
    "list",
    "locate",
    "locatebiome",
    "loot",
    "me",
    "msg",
    "particle",
    "placefeature",
    "playsound",
    "recipe",
    "reload",
    "say",
    "schedule",
    "scoreboard",
    "seed",
    "setblock",
    "setworldspawn",
    "spawnpoint",
    "spectate",
    "spreadplayers",
    "stopsound",
    "summon",
    "tag",
    "team",
    "teammsg",
    "teleport",
    "tell",
    "tellraw",
    "time",
    "title",
    "tm",
    "tp",
    "trigger",
    "w",
    "weather",
    "whitelist",
    "worldborder",
    "xp"
    # OP Level 3-4
    "jfr",
    "perf",
    "publish",
    "save-all",
    "save-off",
    "save-on",
    "stop",
    "ban",
    "ban-ip",
    "banlist",
    "debug",
    "deop",
    "kick",
    "op",
    "pardon",
    "pardon-ip",
    "setidletimeout",
    "whitelist"
}
"""All vanilla command (first argument only)"""
